import pytest
import requests_mock

from valr_python import Client


@pytest.fixture
def sync_client():
    return Client()


@pytest.fixture
def sync_client_with_auth():
    return Client(api_key='api_key', api_secret='api_secret')


@pytest.fixture
def mock_sync_client(sync_client):
    sync_client.base_url = 'mock://test/'
    return sync_client


@pytest.fixture
def rest_sync_mocker():
    with requests_mock.Mocker() as m:
        yield m


@pytest.fixture
def rest_sync_mock_resp():
    return {}


@pytest.fixture
def btc_zar():
    return 'BTCZAR'


@pytest.fixture
def btc():
    return 'BTC'


@pytest.fixture
def zar():
    return 'ZAR'



__version__ = '0.2.0'
__author__ = 'Jonathan Els'

import decimal
import logging

from valr_python.rest_client import *  # noqa
from valr_python.ws_client import *  # noqa

__all__ = (rest_client.__all__, ws_client.__all__)


logging.getLogger(__name__).addHandler(logging.NullHandler())

decimal.getcontext().prec = 8
decimal.DefaultContext.prec = 8



import json
import warnings
from json.decoder import JSONDecodeError
from time import sleep
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from requests.exceptions import HTTPError

from valr_python.exceptions import IncompleteOrderWarning
from valr_python.exceptions import RESTAPIException
from valr_python.exceptions import TooManyRequestsWarning
from valr_python.rest_base import MethodClientABC
from valr_python.utils import DecimalEncoder
from valr_python.utils import _get_valr_headers

__all__ = ('Client',)


class Client(MethodClientABC):
    """
        Synchronous Python SDK for the VALR REST API.

            >>> from valr_python import Client
            >>> from valr_python.exceptions import IncompleteOrderWarning
            >>>
            >>> c = Client(api_key='api_key', api_secret='api_secret')
            >>> c.rate_limiting_support = True # honour HTTP 429 "Retry-After" header values
            >>> limit_order = {
            ...     "side": "SELL",
            ...     "quantity": 0.1,
            ...     "price": 10000,
            ...     "pair": "BTCZAR",
            ...     "post_only": True,
            ... }
            >>> try:
            ...    res = c.post_limit_order(**limit_order)
            ...    order_id = res['id']
            ...    print(order_id)
            ... except IncompleteOrderWarning as w:  # HTTP 202 Accepted handling for incomplete orders
            ...    order_id = w.data['id']
            ...    print(order_id)
            ... except Exception as e:
            ...    print(e)
            "558f5e0a-ffd1-46dd-8fae-763d93fa2f25"
            >>>
        """

    def _do(self, method: str, path: str, data: Optional[Dict] = None,
            is_authenticated: bool = False) -> Optional[Union[List, Dict]]:
        """Executes API request and returns the response.

        Includes HTTP 429 handling by honouring VALR's 429 Retry-After header cool-down.
        """
        headers = {}
        if data:
            data = json.dumps(data, cls=DecimalEncoder)  # serialize decimals as str
            headers["Content-Type"] = "application/json"
        if is_authenticated:
            headers.update(_get_valr_headers(api_key=self.api_key, api_secret=self.api_secret, method=method,
                                             path=path, data=data))
        url = self._base_url + '/' + path.lstrip('/')
        args = dict(timeout=self._timeout, data=data, headers=headers)
        res = self._session.request(method, url, **args)

        try:
            res.raise_for_status()
            e = res.json()
            self._raise_for_api_error(e)
            # provide warning with bundled response dict for incomplete transactions
            if res.status_code == 202:
                warnings.warn(IncompleteOrderWarning(data=e, message="Order processing incomplete"))
            return e
        except HTTPError as he:
            print(he)
            if res.status_code == 429:
                if self._rate_limiting_support:
                    try:
                        retry_after = float(res.headers['Retry-After'])
                        warnings.warn(f"HTTP 429 response received. Applying Retry-After {retry_after}sec back-off",
                                      TooManyRequestsWarning)
                        sleep(retry_after)
                        return self._do(method=method, path=path, is_authenticated=is_authenticated, data=data)
                    except (KeyError, ValueError):
                        raise RESTAPIException(res.status_code,
                                               f'valr-python: HTTP 429 processing failed. '
                                               f'HTTP ({res.status_code}): {res.headers}')
                else:
                    # avoid JSONDecodeError - VALR 429 response has html body
                    raise he
            e = res.json()
            self._raise_for_api_error(e)
            # bubble HTTP errors that VALR API doesn't report on
            raise he
        except JSONDecodeError as jde:
            raise RESTAPIException(res.status_code,
                                   f'valr-python: unknown API error. HTTP ({res.status_code}): {jde.msg}')



import json

from valr_python.utils import _sign_request


def test_request_signature_basic(sync_client):
    sync_client.api_secret = 'superdupersecret'
    timestamp = 1577572690093
    method = 'DELETE'
    path = '/v1/orders/order'
    data = {
        "orderId": "UUID",
        "pair": "BTCZAR"
    }
    body = json.loads(json.dumps(data))
    signature = _sign_request(api_secret=sync_client.api_secret, timestamp=timestamp,
                              method=method, path=path, body=body)
    assert signature == '862ab6527f1ec72bb2243e5f01ae66515d0e74ef1e36aa68c031c045df1b3b62bd43858642a8425368895354e360715add8e3aec47432ea69f60bf6cbd546ea5'  # noqa


def test_request_signature_empty_body(sync_client):
    sync_client.api_secret = 'superdupersecret'
    timestamp = 1577572690093
    method = 'GET'
    path = '/v1/account/balances'
    body = ''
    signature = _sign_request(api_secret=sync_client.api_secret, timestamp=timestamp,
                              method=method, path=path, body=body)
    assert signature == '647d276537b952fe37f349422a4a60a76ecc2e3fad509a523b03dccd1a940525f8ff06314ad1adc5625000223c514637cd9682ee89ffc285b7493e7c64e746aa'  # noqa



from abc import ABCMeta
from abc import abstractmethod
from decimal import Decimal
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import requests

from valr_python.decorators import check_xor_attrs
from valr_python.decorators import requires_authentication
from valr_python.exceptions import APIError

__all__ = ()

DEFAULT_TIMEOUT = 10


class BaseClientABC(metaclass=ABCMeta):
    VALR_API_URL = 'https://api.valr.com'

    def __init__(self, api_key: str = "", api_secret: str = "", timeout: int = 10, base_url: str = "",
                 rate_limiting_support: bool = False) -> None:
        self._api_key = api_key
        self._api_secret = api_secret
        self._base_url = base_url.rstrip('/') if base_url else self.VALR_API_URL
        self._timeout = self.check_timeout(timeout)
        self._rate_limiting_support = rate_limiting_support
        self._session = requests.Session()

    @property
    def api_key(self) -> str:
        return self._api_key

    @api_key.setter
    def api_key(self, value: str) -> None:
        self._api_key = value

    @property
    def api_secret(self) -> str:
        return self._api_secret

    @api_secret.setter
    def api_secret(self, value: str) -> None:
        self._api_secret = value

    @property
    def timeout(self) -> int:
        return self._timeout

    @timeout.setter
    def timeout(self, value: int) -> None:
        self._timeout = self.check_timeout(value)

    @property
    def base_url(self) -> str:
        return self._base_url

    @base_url.setter
    def base_url(self, value: str) -> None:
        self._base_url = value.rstrip('/') if value else self.VALR_API_URL

    @property
    def rate_limiting_support(self) -> bool:
        return self._rate_limiting_support

    @rate_limiting_support.setter
    def rate_limiting_support(self, value: bool) -> None:
        self._rate_limiting_support = value

    @staticmethod
    def check_timeout(timeout: int) -> int:
        """Check if request is non-zero and set to 10 if zero. """
        if timeout == 0:
            return DEFAULT_TIMEOUT
        return timeout

    @staticmethod
    def _raise_for_api_error(e):
        """Raise api responses containing error codes"""
        if 'code' in e and 'message' in e:
            raise APIError(e['code'], e['message'])

    @abstractmethod
    def _do(self, method: str, path: str, data: Dict = None,
            is_authenticated: bool = False) -> Optional[Union[List, Dict]]:
        """Executes API request and returns the response."""
        raise NotImplementedError


class MethodClientABC(BaseClientABC, metaclass=ABCMeta):

    @abstractmethod
    def _do(self, method: str, path: str, data: Dict = None,
            is_authenticated: bool = False) -> Optional[Union[List, Dict]]:
        """Executes API request and returns the response."""
        raise NotImplementedError

    # Public APIs

    def get_order_book_public(self, currency_pair: str) -> Dict[str, List]:
        """Makes a call to GET https://api.valr.com/v1/public/:currencyPair/orderbook

        Returns a list of the top 20 bids and asks in the order book.
        Ask orders are sorted by price ascending.
        Bid orders are sorted by price descending. Orders of the same price are aggregated.

        Please note: This is not an authenticated call.
        More constrained rate-limiting rules will apply than when you use :currencyPair/orderbook route.
        """
        return self._do('GET', f'/v1/public/{currency_pair}/orderbook')

    def get_currencies(self) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/public/currencies

        Get a list of currencies supported by VALR.
        """
        return self._do('GET', '/v1/public/currencies')

    def get_currency_pairs(self) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/public/pairs

        Get a list of all the currency pairs supported by VALR.
        """
        return self._do('GET', '/v1/public/pairs')

    def get_order_types(self, currency_pair: str = "") -> Union[List[Dict], List[str]]:
        """Makes a call to GET https://api.valr.com/v1/public/ordertypes

        Get all the order types supported for all currency pairs.

        An array of currency pairs is returned along with an array of order types for each currency pair.
        You can only place an order that is supported by that currency pair.


        OR


        Makes a call to GET https://api.valr.com/v1/public/:currencyPair/ordertypes

        Get the order types supported for a given currency pair.

        An array of order types is returned. You can only place an order that is listed in this
        array for this currency pair.
        """
        if currency_pair:
            return self._do('GET', f'/v1/public/{currency_pair}/ordertypes')
        else:
            return self._do('GET', '/v1/public/ordertypes')

    def get_market_summary(self, currency_pair: str = "") -> Union[List[Dict], Dict]:
        """Makes a call to GET https://api.valr.com/v1/public/marketsummary

        Get the market summary for all supported currency pairs.


        OR


        Makes a call to GET https://api.valr.com/v1/public/:currencyPair/marketsummary

        Get the market summary for a given currency pair.
        """
        if currency_pair:
            return self._do('GET', f'/v1/public/{currency_pair}/marketsummary')
        else:
            return self._do('GET', '/v1/public/marketsummary')

    def get_server_time(self) -> Dict:
        """Makes a call to GET https://api.valr.com/v1/public/time

        Get the server time. Please note: The server time is returned in seconds.
        """
        return self._do('GET', '/v1/public/time')

    # Account APIs

    @requires_authentication
    def get_balances(self) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/account/balances

        Returns the list of all wallets with their respective balances.
        """
        return self._do('GET', '/v1/account/balances', is_authenticated=True)

    @requires_authentication
    def get_transaction_history(self, skip: int = 0, limit: int = 100) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/account/transactionhistory?skip=0&limit=100

        Transaction history for your account. Note: This API supports pagination.
        """
        return self._do('GET', f'/v1/account/transactionhistory?skip={skip}&limit={limit}', is_authenticated=True)

    @requires_authentication
    def get_account_trade_history(self, currency_pair: str, limit: int = 10) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/account/:currencyPair/tradehistory?limit=10

        Get the last 100 recent trades for a given currency pair for your account.
        You can limit the number of trades returned by specifying the `limit` parameter.
        """
        return self._do('GET', f'/v1/account/{currency_pair}/tradehistory?limit={limit}', is_authenticated=True)

    # Crypto Wallet APIs

    @requires_authentication
    def get_deposit_address(self, currency_code: str) -> Dict:
        """Makes a call to GET https://api.valr.com/v1/wallet/crypto/:currencyCode/deposit/address

        Returns the default deposit address associated with currency specified in the path variable `:currencyCode`.
        """
        return self._do('GET', f'/v1/wallet/crypto/{currency_code}/deposit/address', is_authenticated=True)

    @requires_authentication
    def get_withdrawal_info(self, currency_code: str) -> Dict:
        """Makes a call to GET https://api.valr.com/v1/wallet/crypto/:currencyCode/withdraw

        Get all the information about withdrawing a given currency from your VALR account.
        That will include withdrawal costs, minimum withdrawal amount etc.
        """
        return self._do('GET', f'/v1/wallet/crypto/{currency_code}/withdraw', is_authenticated=True)

    @requires_authentication
    def post_new_crypto_withdrawal(self, currency_code: str, amount: Decimal, address: str,
                                   payment_reference: str = "") -> Dict:
        """Makes a call to POST https://api.valr.com/v1/wallet/crypto/:currencyCode/withdraw

        Withdraw cryptocurrency funds to an address.

        The request body for XRP, XMR, XEM, XLM will accept an optional field called "paymentReference".
        Max length for paymentReference is 256.
        """
        data = {"amount": amount, "address": address}
        if payment_reference:
            data["paymentReference"] = payment_reference
        return self._do('POST', f'/v1/wallet/crypto/{currency_code}/withdraw', data=data, is_authenticated=True)

    @requires_authentication
    def get_withdrawal_status(self, currency_code: str, withdraw_id: str) -> Dict:
        """Makes a call to GET https://api.valr.com/v1/wallet/crypto/:currencyCode/withdraw/:withdrawId

        Check the status of a withdrawal.
        """
        return self._do('GET', f'/v1/wallet/crypto/{currency_code}/withdraw/{withdraw_id}', is_authenticated=True)

    @requires_authentication
    def get_deposit_history(self, currency_code: str, skip: int = 0, limit: int = 10) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/wallet/crypto/:currencyCode/deposit/history?skip=0&limit=10

        Get the Deposit History records for a given currency.
        """
        return self._do('GET', f'/v1/wallet/crypto/{currency_code}/deposit/history?skip={skip}&limit={limit}',
                        is_authenticated=True)

    @requires_authentication
    def get_withdrawal_history(self, currency_code: str, skip: int = 0, limit: int = 10) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/wallet/crypto/:currencyCode/withdraw/history?skip=0&limit=10

        Get Withdrawal History records for a given currency.
        """
        return self._do('GET', f'/v1/wallet/crypto/{currency_code}/withdraw/history?skip={skip}&limit={limit}',
                        is_authenticated=True)

    # Fiat Wallet APIs

    @requires_authentication
    def get_bank_accounts(self, currency_code: str) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/wallet/fiat/:currencyCode/accounts

        Get a list of bank accounts that are linked to your VALR account.
        Bank accounts can be linked by signing in to your account on www.VALR.com.

        :param currency_code: The currency code for the fiat currency. Supported: ZAR.
        :return: list of bank accounts
        """
        return self._do('GET', f'/v1/wallet/fiat/{currency_code}/accounts', is_authenticated=True)

    @requires_authentication
    def post_new_fiat_withdrawal(self, currency_code: str, linked_bank_account_id: str, amount: Decimal) -> Dict:
        """Makes a call to POST https://api.valr.com/v1/wallet/fiat/:currencyCode/withdraw

        Withdraw your ZAR funds into one of your linked bank accounts.
        """
        data = {"linkedBankAccountId": linked_bank_account_id, "amount": amount}
        return self._do('POST', f'/v1/wallet/fiat/{currency_code}/withdraw', data=data, is_authenticated=True)

    # Market Data APIs

    @requires_authentication
    def get_order_book(self, currency_pair: str) -> Dict[str, List]:
        """Makes a call to GET https://api.valr.com/v1/marketdata/:currencyPair/orderbook

        Returns a list of the top 20 bids and asks in the order book.
        Ask orders are sorted by price ascending.
        Bid orders are sorted by price descending. Orders of the same price are aggregated.
        """
        return self._do('GET', f'/v1/marketdata/{currency_pair}/orderbook', is_authenticated=True)

    @requires_authentication
    def get_order_book_full(self, currency_pair: str) -> Dict[str, List]:
        """Makes a call to GET https://api.valr.com/v1/marketdata/:currencyPair/orderbook/full

        Returns a list of all the bids and asks in the order book.
        Ask orders are sorted by price ascending.
        Bid orders are sorted by price descending. Orders of the same price are NOT aggregated..
        """
        return self._do('GET', f'/v1/marketdata/{currency_pair}/orderbook/full', is_authenticated=True)

    @requires_authentication
    def get_market_data_trade_history(self, currency_pair: str, limit: int = 10) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/marketdata/:currencyPair/tradehistory?limit=10

        Get the last 100 recent trades for a given currency pair.
        You can limit the number of trades returned by specifying the limit parameter.
        """
        return self._do('GET', f'/v1/marketdata/{currency_pair}/tradehistory?limit={limit}', is_authenticated=True)

    # Simple Buy/Sell APIs

    @requires_authentication
    def post_simple_quote(self, currency_pair: str, pay_in_currency: str, pay_amount: Decimal, side: str) -> Dict:
        """Makes a call to POST https://api.valr.com/v1/simple/:currencyPair/quote

        Get a quote to buy or sell instantly using Simple Buy.

        A sample request body is as follows:

        {
            "payInCurrency": "BTC",
            "payAmount": "0.001",
            "side": "SELL"
        }
        Example usage of payInCurrency and side for the BTCZAR currency pair:

         - If you want to sell BTC for ZAR, payInCurrency will be BTC and the side would be SELL
         - If you want to buy BTC with ZAR, payInCurrency will be ZAR and the side would be BUY

        Example usage of payInCurrency and side for the ETHBTC currency pair:

         - If you want to sell ETH for BTC, payInCurrency will be ETH and the side would be SELL
         - If you want to buy ETH with BTC, payInCurrency will be BTC and the side would be BUY
        """
        data = {"payInCurrency": pay_in_currency, "payAmount": pay_amount, "side": side}
        return self._do('POST', f'/v1/simple/{currency_pair}/quote', data=data, is_authenticated=True)

    @requires_authentication
    def post_simple_order(self, currency_pair: str, pay_in_currency: str, pay_amount: Decimal, side: str) -> Dict:
        """Makes a call to POST https://api.valr.com/v1/simple/:currencyPair/order

        Submit an order to buy or sell instantly using Simple Buy/Sell.

        A sample request body is as follows:

        {
            "payInCurrency": "BTC",
            "payAmount": "0.001",
            "side": "SELL"
        }
        Example usage of payInCurrency and side for the BTCZAR currency pair:

         - If you want to sell BTC for ZAR, payInCurrency will be BTC and the side would be SELL
         - If you want to buy BTC with ZAR, payInCurrency will be ZAR and the side would be BUY

        Example usage of payInCurrency and side for the ETHBTC currency pair:

         - If you want to sell ETH for BTC, payInCurrency will be ETH and the side would be SELL
         - If you want to buy ETH with BTC, payInCurrency will be BTC and the side would be BUY
        """
        data = {"payInCurrency": pay_in_currency, "payAmount": pay_amount, "side": side}
        return self._do('POST', f'/v1/simple/{currency_pair}/order', data=data, is_authenticated=True)

    @requires_authentication
    def get_simple_order_status(self, currency_pair: str, order_id: str) -> Dict:
        """Makes a call to GET https://api.valr.com/v1/simple/:currencyPair/order/:orderId

        Get the status of a Simple Buy/Sell order.
        """
        return self._do('GET', f'/v1/simple/{currency_pair}/order/{order_id}', is_authenticated=True)

    # Exchange Buy/Sell APIs

    @requires_authentication
    def post_limit_order(self, side: str, quantity: Decimal, price: Decimal, pair: str, post_only: bool = False,
                         customer_order_id: str = "") -> Dict:
        """Makes a call to POST https://api.valr.com/v1/orders/limit

        Create a new limit order.

        The JSON body used to create a limit order looks like this:

        {
            "side": "SELL",
            "quantity": "0.100000",
            "price": "10000",
            "pair": "BTCZAR",
            "postOnly": true,
            "customerOrderId": "1234"
        }


        side (required)	- BUY or SELL
        quantity (required)	- ase amount in BTC
        price (required) - Price per coin in ZAR
        pair (required) - BTCZAR
        postOnly (optional)	- true or false
        customerOrderId (optional) - Numeric value. See below for explanation.

        The customerOrderId is an optional field which can be specified by clients to track
        this order using their own internal order management systems.

        - When you retrieve your open orders or history of orders, this field will also be returned along with the
        VALR's orderId for each order for which this field was specified during creation.
        - When you cancel an order, you can either specify orderId or customerOrderId, not both.
        - customerOrderId is alphanumeric with no special chars, limit of 50 characters.
        - The customerOrderId has to be unique accross all open orders for a given account.
        If you do reuse an id value that is currently an active open order, your order will not be placed
        (you can check the status of an order using the order status API call).

        Fee currency

        - If you are a Maker and you are Buying BTC with ZAR, your reward will be paid in ZAR
        - If you are a Maker and you are Selling BTC for ZAR, your reward will be paid in BTC
        - If you are a Taker and you are Buying BTC with ZAR, your fee will be charged in BTC
        - If you are a Taker and you are Selling BTC for ZAR, your fee will be charged in ZAR

        In short, fees will be in the currency that the taker is receiving and the maker is paying in the trade.

        BTC/ZAR	    MAKER	TAKER	FEE/REWARD CURRENCY
        TRADE SIDE	BUY	    SELL	ZAR
        TRADE SIDE	SELL	BUY	    BTC

        PLEASE NOTE: When you receive a response with an id, it does not always mean that the order has been placed.
        When the response is 202 Accepted, you can use the Order Status REST API or use WebSocket API
        to receive the status of this order. The reasons why an order could fail are as follows:

        - Insufficient balance in your account.
        - If you're using a non-unique customerOrderId.
        - If you set postOnly flag to true, but your order would have matched immediately.
        - Self trading: If your order matches against your own order (on the other side).
        - Insufficient liquidity: If you're placing an order and there isn't liquidity to fulfill the order.
        """
        data = {
            "side": side,
            "quantity": quantity,
            "price": price,
            "pair": pair
        }
        if post_only:
            data["postOnly"] = post_only
        if customer_order_id:
            data["customerOrderId"] = customer_order_id
        return self._do('POST', '/v1/orders/limit', data=data, is_authenticated=True)

    @requires_authentication
    @check_xor_attrs("base_amount", "quote_amount")
    def post_market_order(self, side: str, pair: str, base_amount: Decimal = None, quote_amount: Decimal = None,
                          customer_order_id: str = "") -> Dict:
        """Makes a call to POST https://api.valr.com/v1/orders/market

        Create a new market order.

        When the response is 202 Accepted, you can either use the Order Status REST API
        or use WebSocket API to receive updates about this order.

        Example request body:

        {
            "side": "SELL",
            "baseAmount": "0.100000",
            "pair": "BTCZAR",
            "customerOrderId": "1234"
        }

        side (required)	- BUY or SELL
        baseAmount / quoteAmount (required)	- Quote amount for BUY (in ZAR). Base amount for SELL (in BTC)
        pair (required)	- BTCZAR
        customerOrderId (optional) - Numeric value. See below for explanation.

        The customerOrderId is an optional field which can be specified by clients to track this order
        using their own internal order management systems.

        - When you retrieve your open orders or history of orders, this field will also be returned along
        with the VALR's orderId for each order for which this field was specified during creation.
        - When you cancel an order, you can either specify orderId or customerOrderId, not both.
        - customerOrderId is alphanumeric with no special chars, limit of 50 characters.
        - The customerOrderId has to be unique accross all open orders for a given account.
        If you do reuse an id value that is currently an active open order, your order will not be placed
        (you can check the status of an order using the order status API call).

        Fee currency

        When you place a market order, you will be charged the "Taker fee" on the trade.

        - As a Taker, if you are Buying BTC with ZAR, your fee will be charged in BTC
        - As a Taker, if you are Selling BTC for ZAR, your fee will be charged in ZAR

        PLEASE NOTE: When you receive a response with an id, it does not always mean that the order has been placed.
        When the response is 202 Accepted, you can use the Order Status REST API or use WebSocket API to receive
        the status of this order. The reasons why an order could fail are as follows:

        - Insufficient balance in your account.
        - If you're using a non-unique customerOrderId.
        - If you set postOnly flag to true, but your order would have matched immediately.
        - Self trading: If your order matches against your own order (on the other side).
        - Insufficient liquidity: If you're placing an order and there isn't liquidity to fulfill the order.
        """
        data = {
            "side": side,
            "pair": pair
        }
        if base_amount:
            data["baseAmount"] = base_amount
        else:
            data["quoteAmount"] = quote_amount
        if customer_order_id:
            data["customerOrderId"] = customer_order_id
        return self._do('POST', '/v1/orders/market', data=data, is_authenticated=True)

    @requires_authentication
    @check_xor_attrs("order_id", "customer_order_id")
    def get_order_status(self, currency_pair: str, order_id: str = "", customer_order_id: str = "") -> Dict:
        """Makes a call to GET https://api.valr.com/v1/orders/:currencyPair/orderid/:orderId

        This API returns the status of an order that was placed on the Exchange queried using the id provided by VALR.
        VALR provides an id for every order that is placed on the Exchange.
        Use this id to populate the path variable orderId in this API to query the status of the order.

        Note: If a customerOrderId was also specified while placing the order,
        that customerOrderId will be returned as part of the response.


        OR


        Makes a call to GET https://api.valr.com/v1/orders/:currencyPair/customerorderid/:customerOrderId

        This API returns the status of an order that was placed on the Exchange queried using customerOrderId.
        The customer can specify a customerOrderId while placing an order on the Exchange.
        Use this API to query the order status using that customerOrderId.
        """
        if customer_order_id:
            return self._do('GET', f'/v1/orders/{currency_pair}/customerorderid/{customer_order_id}',
                            is_authenticated=True)
        else:
            return self._do('GET', f'/v1/orders/{currency_pair}/orderid/{order_id}', is_authenticated=True)

    @requires_authentication
    def get_all_open_orders(self) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/orders/open

        Get all open orders for your account.

        A customerOrderId field will be returned in the response for all those orders
        that were created with a customerOrderId field.
        """
        return self._do('GET', '/v1/orders/open', is_authenticated=True)

    @requires_authentication
    def get_order_history(self, skip: int = 0, limit: int = 2) -> List[Dict]:
        """Makes a call to GET https://api.valr.com/v1/orders/history?skip=0&limit=2

        Get historical orders placed by you.
        """
        return self._do('GET', f'/v1/orders/history?skip={skip}&limit={limit}', is_authenticated=True)

    @requires_authentication
    @check_xor_attrs("order_id", "customer_order_id")
    def get_order_history_summary(self, order_id: str = "", customer_order_id: str = "") -> Dict:
        """Makes a call to GET https://api.valr.com/v1/orders/history/summary/orderid/:orderId

        An order is considered completed when the "Order Status" call returns one of the following statuses:

        "Filled", "Cancelled" or "Failed".

        When this happens, you can get a more detailed summary about this order using this call.
        Orders that are not completed are invalid for this request.


        OR


        Makes a call to GET https://api.valr.com/v1/orders/history/summary/customerorderid/:customerOrderId

        An order is considered completed when the "Order Status" call returns one of the following statuses:

        "Filled", "Cancelled" or "Failed".

        When this happens, you can get a more detailed summary about this order using this call.
        Orders that are not completed are invalid for this request.
        """
        if customer_order_id:
            return self._do('GET', f'/v1/orders/history/summary/customerorderid/{customer_order_id}',
                            is_authenticated=True)
        else:
            return self._do('GET', f'/v1/orders/history/summary/orderid/{order_id}', is_authenticated=True)

    @requires_authentication
    @check_xor_attrs("order_id", "customer_order_id")
    def get_order_history_detail(self, order_id: str = "", customer_order_id: str = "") -> Dict:
        """Makes a call to GET https://api.valr.com/v1/orders/history/detail/orderid/:orderId

        Get a detailed history of an order's statuses. This call returns an array of "Order Status" objects.
        The latest and most up-to-date status of this order is the zeroth element in the array.


        OR


        Makes a call to GET https://api.valr.com/v1/orders/history/detail/customerorderid/:customerOrderId

        Get a detailed history of an order's statuses. This call returns an array of "Order Status" objects.
        The latest and most up-to-date status of this order is the zeroth element in the array.
        """
        if customer_order_id:
            return self._do('GET', f'/v1/orders/history/detail/customerorderid/{customer_order_id}',
                            is_authenticated=True)
        else:
            return self._do('GET', f'/v1/orders/history/detail/orderid/{order_id}', is_authenticated=True)

    @requires_authentication
    @check_xor_attrs("order_id", "customer_order_id")
    def delete_order(self, pair, order_id: str = "", customer_order_id: str = "") -> None:
        """Makes a call to DELETE https://api.valr.com/v1/orders/order

        Cancel an open order.

        A 202 Accepted response means the request to cancel the order was accepted.
        You can either use the Order Status REST API or use WebSocket API to receive status update about this request.

        The DELETE request requires a JSON request body in the following format:

        {
          "orderId": "UUID",
          "pair": "BTCZAR"
        }
        Alternatively, the body can be of the following format if you specified a customerOrderId
        when creating your order:

        {
          "customerOrderId": "^[0-9a-zA-Z-]{0,50}$",
          "pair": "BTCZAR"
        }
        NOTE: When you receive this response with an id, it does not always mean that the order has been placed.
        When the response is 202 Accepted, you can either use the Order Status REST API
        or use WebSocket API to receive status update about this order.
        """
        data = {"pair": pair}
        if order_id:
            data["orderId"] = order_id
        else:
            data["customerOrderId"] = customer_order_id
        return self._do('DELETE', '/v1/orders/order', data=data, is_authenticated=True)



from functools import wraps
from typing import List

from valr_python.exceptions import RequiresAuthentication

__all__ = ()


def requires_authentication(func):
    """Decorator to determine private API calls with require authentication"""

    @wraps(func)
    def inner(self, *args, **kwargs):
        if not (self._api_key and self._api_secret):
            raise RequiresAuthentication("cannot generate private request without API key/secret.")
        return func(self, *args, **kwargs)

    return inner


def check_xor_attrs(*xor_args: List[str]):
    """Decorator to check that only one of two attributes was provided in function kwargs"""

    def xor_decorator(func):

        @wraps(func)
        def inner(self, *args, **kwargs):
            if len(xor_args) != 2:
                raise AttributeError("only comparisons of two args supported")
            has_args = [i in kwargs for i in xor_args]
            if not any(has_args) or all(has_args):
                raise AttributeError(f"either {xor_args[0]} or {xor_args[1]} must be provided, but not both.")
            return func(self, *args, **kwargs)

        return inner

    return xor_decorator



__all__ = (
    'APIError',
    'RESTAPIException',
    'RequiresAuthentication',
    'TooManyRequestsWarning',
    'WebSocketAPIException',
    'IncompleteOrderWarning',
    'HookNotFoundError'
)


class APIError(Exception):
    """API error responses"""
    def __init__(self, code, message):
        self.code = code
        self.message = message


class RESTAPIException(Exception):
    """Unhandled Rest API exceptions"""
    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message


class WebSocketAPIException(Exception):
    """Unhandled Websocket API exceptions"""


class RequiresAuthentication(Exception):
    """Request requires API authentication"""


class TooManyRequestsWarning(Warning):
    """HTTP 429 received and handled"""


class IncompleteOrderWarning(Warning):
    """HTTP 202 Accepted response received for incomplete order processing"""
    def __init__(self, message, data):
        self.message = message
        self.data = data


class HookNotFoundError(Exception):
    """Could not map websocket event from supplied hooks"""



from enum import Enum
from enum import auto

__all__ = (
    'WebSocketType',
    'MessageFeedType',
    'TradeEvent',
    'AccountEvent',
    'OrderStatusType',
    'CurrencyPair'
)


class WebSocketType(Enum):
    ACCOUNT = '/ws/account'
    TRADE = '/ws/trade'


class MessageFeedType(Enum):
    SUBSCRIBE = auto()
    SUBSCRIBED = auto()
    AUTHENTICATED = auto()
    UNSUPPORTED = auto()
    PING = auto()
    PONG = auto()


class TradeEvent(Enum):
    AGGREGATED_ORDERBOOK_UPDATE = auto()
    MARKET_SUMMARY_UPDATE = auto()
    NEW_TRADE_BUCKET = auto()
    NEW_TRADE = auto()


class AccountEvent(Enum):
    NEW_ACCOUNT_HISTORY_RECORD = auto()
    BALANCE_UPDATE = auto()
    NEW_ACCOUNT_TRADE = auto()
    INSTANT_ORDER_COMPLETED = auto()
    OPEN_ORDERS_UPDATE = auto()
    ORDER_PROCESSED = auto()
    ORDER_STATUS_UPDATE = auto()
    FAILED_CANCEL_ORDER = auto()
    NEW_PENDING_RECEIVE = auto()
    SEND_STATUS_UPDATE = auto()


class OrderStatusType(Enum):
    PLACED = "Placed"
    FAILED = "Failed"
    CANCELLED = "Cancelled"
    FILLED = "Filled"
    PARTIALLY_FILLED = "Partially Filled"
    INSTANT_ORDER_BALANCE_RESERVE_FAILED = "Instant Order Balance Reserve Failed"
    INSTANT_ORDER_BALANCE_RESERVED = "Instant Order Balance Reserved"
    INSTANT_ORDER_COMPLETED = "Instant Order Completed"


class CurrencyPair(Enum):
    BTCZAR = auto()
    ETHZAR = auto()



import decimal
import hashlib
import hmac
import json
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from valr_python.enum import WebSocketType
from valr_python.exceptions import RequiresAuthentication

__all__ = ()

JSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


def _get_valr_headers(api_key: str, api_secret: str, method: str, path: Union[str, WebSocketType],
                      data: str) -> Dict:
    """Create signed VALR headers from method, api path and request params

    :param method: HTTP method (e.g. GET, POST, DELETE, etc.
    :param path: REST API endpoint path
    :param data: params dict for request body
    :return: header dict
    """
    valr_headers = {}
    if not (api_key and api_secret):
        raise RequiresAuthentication("Cannot generate private request without API key/secret.")
    timestamp = int(time.time() * 1000)
    valr_headers["X-VALR-API-KEY"] = api_key
    valr_headers["X-VALR-SIGNATURE"] = _sign_request(api_secret=api_secret, timestamp=timestamp,
                                                     method=method, path=path,
                                                     body=data)
    valr_headers["X-VALR-TIMESTAMP"] = str(timestamp)  # str or byte req for request headers

    return valr_headers


def _sign_request(api_secret: str, timestamp: int, method: str, path: str, body: str = "") -> str:
    """Signs the request payload using the api key secret

    :param timestamp: the unix timestamp of this request e.g. int(time.time()*1000)
    :param method: Http method - GET, POST, PUT or DELETE
    :param path: path excluding FQDN
    :param body: http request body as a JSON string, optional
    :return signature hash
    """
    body = body if body else ""
    payload = f"{timestamp}{method.upper()}{path}{body}"
    message = bytearray(payload, 'utf-8')
    signature = hmac.new(bytearray(api_secret, 'utf-8'), message, digestmod=hashlib.sha512).hexdigest()
    return signature


class DecimalEncoder(json.JSONEncoder):
    """Serialize Decimal obj as str"""
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return str(o)
        return super(DecimalEncoder, self).default(o)



import asyncio
import json
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import websockets

from valr_python.enum import AccountEvent
from valr_python.enum import CurrencyPair
from valr_python.enum import MessageFeedType
from valr_python.enum import TradeEvent
from valr_python.enum import WebSocketType
from valr_python.exceptions import HookNotFoundError
from valr_python.exceptions import WebSocketAPIException
from valr_python.utils import JSONType
from valr_python.utils import _get_valr_headers

__all__ = ('WebSocketClient',)


def get_event_type(ws_type: WebSocketType) -> Type[Union[TradeEvent, AccountEvent]]:
    return TradeEvent if ws_type == WebSocketType.TRADE else AccountEvent


class WebSocketClient:
    """The WebSocket API is an advanced technology that makes it possible to open a two-way interactive
    communication session between a client and a server. With this API, you can send messages to a server and
    receive event-driven responses without having to poll the server for a reply.


    Example Usage
    ~~~~~~~~~~~~~

    >>> import asyncio
    >>> from typing import Dict
    >>> from pprint import pprint
    >>> from valr_python import WebSocketClient
    >>> from valr_python.enum import TradeEvent
    >>> from valr_python.enum import WebSocketType
    >>>
    >>> def pretty_hook(data: Dict):
    ...    pprint(data)
    >>>
    >>> c = WebSocketClient(api_key='api_key', api_secret='api_secret', currency_pairs=['BTCZAR'],
    ...                     ws_type=WebSocketType.TRADE.name,
    ...                     trade_subscriptions=[TradeEvent.MARKET_SUMMARY_UPDATE.name],
    ...                     hooks={TradeEvent.MARKET_SUMMARY_UPDATE.name : pretty_hook})
    >>> loop = asyncio.get_event_loop()
    >>> loop.run_until_complete(c.run())
    {'currencyPairSymbol': 'BTCZAR',
     'data': {'askPrice': '151601',
              'baseVolume': '314.7631144',
              'bidPrice': '151600',
              'changeFromPrevious': '2.14',
              'created': '2020-02-06T22:47:03.129Z',
              'currencyPairSymbol': 'BTCZAR',
              'highPrice': '152440',
              'lastTradedPrice': '151600',
              'lowPrice': '146765',
              'previousClosePrice': '148410',
              'quoteVolume': '47167382.04552981'},
     'type': 'MARKET_SUMMARY_UPDATE'}

    Connection
    ~~~~~~~~~~

    Our WebSocket API is accessible on the following address: wss://api.valr.com.

    Account WebSocket connection: In order to receive streaming updates about your VALR account, you would
    open up a WebSocket connection to wss://api.valr.com/ws/account

    Trade WebSocket connection: In order to receive streaming updates about Trade data, you would open up a
    WebSocket connection to wss://api.valr.com/ws/trade


    Authentication
    ~~~~~~~~~~~~~~

    Our WebSocket API needs authentication. To authenticate, pass in the following headers to the first
    call that establishes the WebSocket connection.

    X-VALR-API-KEY: Your API Key
    X-VALR-SIGNATURE: Generated signature. The signature is generated using the following parameters:
        Api Secret
        Timestamp of request
        HTTP verb 'GET'
        Path (either /ws/account or /ws/trade)
        Request Body should be empty
    X-VALR-TIMESTAMP: Timestamp of the request

    The headers that are passed to establish the connection are the same 3 headers you pass to
    any authenticated call to the REST API.


    Subscribing to events
    ~~~~~~~~~~~~~~~~~~~~~

    Once you open a connection to Account, you are automatically subscribed to all messages for all events on
    the Account WebSocket connection. You will start receiving message feeds pertaining to your VALR account.
    For example, you will receive messages when your balance is updated or when a new trade is executed on your account.

    On the other hand, when you open a connection to Trade, in order to receive message feeds about trading data, you
    must subscribe to events you are interested in on the Trade WebSocket connection.  For example, if you want to
    receive messages when markets fluctuate, you must send a message on the connection with the following payload:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"MARKET_SUMMARY_UPDATE",
             "pairs":[
                "BTCZAR"
             ]
          }
       ]
    }

    Here, the event you are subscribing to is called MARKET_SUMMARY_UPDATE and the currency pair you are subscribing to
     is an array. We currently only support BTCZAR and ETHZAR. XRPZAR will be added in due course.


    Unsubscribing from events
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    When you are no longer interested in receiving messages for certain events on the Trade WebSocket connection,
    you can send a synthetic "unsubscribe" message. For example, if you want to unsubscribe from MARKET_SUMMARY_UPDATE
    event, you would send a message as follows:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"MARKET_SUMMARY_UPDATE",
             "pairs":[

             ]
          }
       ]
    }


    Staying connected with Ping-Pong messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    To ensure that you stay connected to either the Account or Trade WebSocket you can send a "PING" message on the
    WebSocket you wish to monitor. VALR will respond with a PONG event. The message must be as follows:

    {
      "type": "PING"
    }

    Events (On Trade WebSocket)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Here is a list of events you can subscribe to on the Trade WebSocket connection:

    Event	                        Description

    AGGREGATED_ORDERBOOK_UPDATE	    When subscribed to this event for a given currency pair, the client receives the
                                    top 20 bids and asks from the order book for that currency pair.
    MARKET_SUMMARY_UPDATE	        When subscribed to this event for a given currency pair, the client receives a
                                    message feed with the latest market summary for that currency pair.
    NEW_TRADE_BUCKET	            When subscribed to this event for a given currency pair, the client receives the
                                    Open, High, Low, Close data valid for the last 60 seconds.
    NEW_TRADE	                    When subscribed to this event for a given currency pair, the client receives
                                    message feeds with the latest trades that are executed for that currency pair.


    AGGREGATED_ORDERBOOK_UPDATE

    In order to subscribe to AGGREGATED_ORDERBOOK_UPDATE for BTCZAR and ETHZAR, you must send the following message
    on the Trade WebSocket connection once it is opened:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"AGGREGATED_ORDERBOOK_UPDATE",
             "pairs":[
                "BTCZAR",
                "ETHZAR"
             ]
          }
       ]
    }

    To unsubscribe, send the following message:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"AGGREGATED_ORDERBOOK_UPDATE",
             "pairs":[

             ]
          }
       ]
    }


    MARKET_SUMMARY_UPDATE

    In order to subscribe to MARKET_SUMMARY_UPDATE for just BTCZAR, you must send the following message on the
    Trade WebSocket connection once it is opened:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"MARKET_SUMMARY_UPDATE",
             "pairs":[
                "BTCZAR"
             ]
          }
       ]
    }

    To unsubscribe, send the following message:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"MARKET_SUMMARY_UPDATE",
             "pairs":[

             ]
          }
       ]
    }


    NEW_TRADE_BUCKET

    In order to subscribe to NEW_TRADE_BUCKET for BTCZAR as well as ETHZAR, you must send the following message on the
    Trade WebSocket connection once it is opened:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"NEW_TRADE_BUCKET",
             "pairs":[
                "BTCZAR",
                "ETHZAR"
             ]
          }
       ]
    }

    To unsubscribe, send the following message:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"NEW_TRADE_BUCKET",
             "pairs":[

             ]
          }
       ]
    }


    NEW_TRADE

    In order to subscribe to NEW_TRADE just for BTCZAR, you must send the following message on the Trade WebSocket
    connection once it is opened:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"NEW_TRADE",
             "pairs":[
                "BTCZAR"
             ]
          }
       ]
    }

    To unsubscribe, send the following message:

    {
       "type":"SUBSCRIBE",
       "subscriptions":[
          {
             "event":"NEW_TRADE",
             "pairs":[

             ]
          }
       ]
    }


    Message Feeds (On Trade WebSocket)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    As and when events occur, the message feeds come through to the Trade WebSocket connection for the events the client
    has subscribed to. You will find an example message feed for each event specified above.


    AGGREGATED_ORDERBOOK_UPDATE

    Sample message feed:

    {
       "type":"AGGREGATED_ORDERBOOK_UPDATE",
       "currencyPairSymbol":"BTCZAR",
       "data":{
          "Asks":[
             {
                "side":"sell",
                "quantity":"0.005",
                "price":"9500",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"sell",
                "quantity":"0.01",
                "price":"9750",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"sell",
                "quantity":"0.643689",
                "price":"10000",
                "currencyPair":"BTCZAR",
                "orderCount":3
             },
             {
                "side":"sell",
                "quantity":"0.2",
                "price":"11606",
                "currencyPair":"BTCZAR",
                "orderCount":2
             },
             {
                "side":"sell",
                "quantity":"0.67713484",
                "price":"14000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"sell",
                "quantity":"1",
                "price":"15000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"sell",
                "quantity":"1",
                "price":"16000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"sell",
                "quantity":"1",
                "price":"17000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"sell",
                "quantity":"1",
                "price":"18000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"sell",
                "quantity":"1",
                "price":"19000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             }
          ],
          "Bids":[
             {
                "side":"buy",
                "quantity":"0.038",
                "price":"9000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8802",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.2",
                "price":"8801",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8800",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8700",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8600",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8500",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8400",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.3",
                "price":"8200",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8100",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"0.1",
                "price":"8000",
                "currencyPair":"BTCZAR",
                "orderCount":1
             },
             {
                "side":"buy",
                "quantity":"1.08027437",
                "price":"1",
                "currencyPair":"BTCZAR",
                "orderCount":3
             }
          ]
       }
    }


    MARKET_SUMMARY_UPDATE

    Sample message feed:

    {
       "type":"MARKET_SUMMARY_UPDATE",
       "currencyPairSymbol":"BTCZAR",
       "data":{
          "currencyPairSymbol":"BTCZAR",
          "askPrice":"9500",
          "bidPrice":"9000",
          "lastTradedPrice":"9500",
          "previousClosePrice":"9000",
          "baseVolume":"0.0551",
          "highPrice":"10000",
          "lowPrice":"9000",
          "created":"2016-04-25T19:41:16.237Z",
          "changeFromPrevious":"5.55"
       }
    }


    NEW_TRADE_BUCKET

    Sample message feed:

    {
       "type":"NEW_TRADE_BUCKET",
       "currencyPairSymbol":"BTCZAR",
       "data":{
          "currencyPairSymbol":"BTCZAR",
          "bucketPeriodInSeconds":60,
          "startTime":"2019-04-25T19:41:00Z",
          "open":"9500",
          "high":"9500",
          "low":"9500",
          "close":"9500",
          "volume":"0"
       }
    }


    NEW_TRADE

    Sample message feed:

    {
       "type":"NEW_TRADE",
       "currencyPairSymbol":"BTCZAR",
       "data":{
          "price":"9500",
          "quantity":"0.001",
          "currencyPair":"BTCZAR",
          "tradedAt":"2019-04-25T19:51:55.393Z",
          "takerSide":"buy"
       }
    }


    Message Feeds (On Account WebSocket)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    As and when events occur, the message feeds come through to the Account WebSocket connection. As mentioned
    previously, the client is automatically subscribed to all events on the Account WebSocket connection as soon as
    the connection is established. That means, the client need not subscribe to events on the Account WebSocket
    connection. That also means that the client cannot unsubscribe from these events.

    Here is a list of events that occur on the Account WebSocket and the corresponding sample message feed:


    NEW_ACCOUNT_HISTORY_RECORD : NEW SUCCESSFUL TRANSACTION

    Sample message feed:

    {
       "type":"NEW_ACCOUNT_HISTORY_RECORD",
       "data":{
          "transactionType":{
             "type":"SIMPLE_BUY",
             "description":"Simple Buy"
          },
          "debitCurrency":{
             "symbol":"R",
             "decimalPlaces":2,
             "isActive":true,
             "shortName":"ZAR",
             "longName":"Rand",
             "supportedWithdrawDecimalPlaces":2
          },
          "debitValue":"10",
          "creditCurrency":{
             "symbol":"BTC",
             "decimalPlaces":8,
             "isActive":true,
             "shortName":"BTC",
             "longName":"Bitcoin",
             "supportedWithdrawDecimalPlaces":8
          },
          "creditValue":"0.00104473",
          "feeCurrency":{
             "symbol":"BTC",
             "decimalPlaces":8,
             "isActive":true,
             "shortName":"BTC",
             "longName":"Bitcoin",
             "supportedWithdrawDecimalPlaces":8
          },
          "feeValue":"0.00000789",
          "eventAt":"2019-04-25T20:36:53.426Z",
          "additionalInfo":{
             "costPerCoin":9500,
             "costPerCoinSymbol":"R",
             "currencyPairSymbol":"BTCZAR"
          }
       }
    }


    BALANCE_UPDATE : BALANCE HAS BEEN UPDATED

    Sample message feed:

    {
       "type":"BALANCE_UPDATE",
       "data":{
          "currency":{
             "symbol":"BTC",
             "decimalPlaces":8,
             "isActive":true,
             "shortName":"BTC",
             "longName":"Bitcoin",
             "supportedWithdrawDecimalPlaces":8
          },
          "available":"0.88738681",
          "reserved":"0.97803484",
          "total":"1.86542165"
       }
    }

    NEW_ACCOUNT_TRADE : NEW TRADE EXECUTED ON YOUR ACCOUNT

    Sample message feed:

    {
       "type":"NEW_ACCOUNT_TRADE",
       "currencyPairSymbol":"BTCZAR",
       "data":{
          "price":"9500",
          "quantity":"0.00105263",
          "currencyPair":"BTCZAR",
          "tradedAt":"2019-04-25T20:36:53.426Z",
          "side":"buy"
       }
    }


    INSTANT_ORDER_COMPLETED: NEW SIMPLE BUY/SELL EXECUTED

    Sample message feed:

    {
       "type":"INSTANT_ORDER_COMPLETED",
       "data":{
          "orderId":"247dc157-bb5b-49af-b476-2f613b780697",
          "success":true,
          "paidAmount":"10",
          "paidCurrency":"R",
          "receivedAmount":"0.00104473",
          "receivedCurrency":"BTC",
          "feeAmount":"0.00000789",
          "feeCurrency":"BTC",
          "orderExecutedAt":"2019-04-25T20:36:53.445"
       }
    }


    OPEN_ORDERS_UPDATE : NEW ORDER ADDED TO OPEN ORDERS

    Sample message feed (all open orders are returned) :

    {
       "type":"OPEN_ORDERS_UPDATE",
       "data":[
          {
             "orderId":"38511e49-a755-4f8f-a2b1-232bae6967dc",
             "side":"sell",
             "remainingQuantity":"0.1",
             "originalPrice":"10000",
             "currencyPair":{
                "id":1,
                "symbol":"BTCZAR",
                "baseCurrency":{
                   "id":2,
                   "symbol":"BTC",
                   "decimalPlaces":8,
                   "isActive":true,
                   "shortName":"BTC",
                   "longName":"Bitcoin",
                   "currencyDecimalPlaces":8,
                   "supportedWithdrawDecimalPlaces":8
                },
                "quoteCurrency":{
                   "id":1,
                   "symbol":"R",
                   "decimalPlaces":2,
                   "isActive":true,
                   "shortName":"ZAR",
                   "longName":"Rand",
                   "currencyDecimalPlaces":2,
                   "supportedWithdrawDecimalPlaces":2
                },
                "shortName":"BTC/ZAR",
                "exchange":"VALR",
                "active":true,
                "minBaseAmount":0.0001,
                "maxBaseAmount":2,
                "minQuoteAmount":10,
                "maxQuoteAmount":100000
             },
             "createdAt":"2019-04-17T19:51:35.776Z",
             "originalQuantity":"0.1",
             "filledPercentage":"0.00",
             "customerOrderId":""
          },
          {
             "orderId":"d1d9f20a-778c-4f4a-98a1-d336da960158",
             "side":"sell",
             "remainingQuantity":"0.1",
             "originalPrice":"10000",
             "currencyPair":{
                "id":1,
                "symbol":"BTCZAR",
                "baseCurrency":{
                   "id":2,
                   "symbol":"BTC",
                   "decimalPlaces":8,
                   "isActive":true,
                   "shortName":"BTC",
                   "longName":"Bitcoin",
                   "currencyDecimalPlaces":8,
                   "supportedWithdrawDecimalPlaces":8
                },
                "quoteCurrency":{
                   "id":1,
                   "symbol":"R",
                   "decimalPlaces":2,
                   "isActive":true,
                   "shortName":"ZAR",
                   "longName":"Rand",
                   "currencyDecimalPlaces":2,
                   "supportedWithdrawDecimalPlaces":2
                },
                "shortName":"BTC/ZAR",
                "exchange":"VALR",
                "active":true,
                "minBaseAmount":0.0001,
                "maxBaseAmount":2,
                "minQuoteAmount":10,
                "maxQuoteAmount":100000
             },
             "createdAt":"2019-04-20T13:48:44.922Z",
             "originalQuantity":"0.1",
             "filledPercentage":"0.00",
             "customerOrderId":"4"
          }
       ]
    }


    ORDER_PROCESSED : ORDER PROCESSED

    Sample message feed:

    {
       "type":"ORDER_PROCESSED",
       "data":{
          "orderId":"247dc157-bb5b-49af-b476-2f613b780697",
          "success":true,
          "failureReason":""
       }
    }


    ORDER_STATUS_UPDATE : ORDER STATUS HAS BEEN UPDATED

    Sample message feed:

    {
       "type":"ORDER_STATUS_UPDATE",
       "data":{
          "orderId":"247dc157-bb5b-49af-b476-2f613b780697",
          "orderStatusType":"Filled",
          "currencyPair":{
             "id":1,
             "symbol":"BTCZAR",
             "baseCurrency":{
                "id":2,
                "symbol":"BTC",
                "decimalPlaces":8,
                "isActive":true,
                "shortName":"BTC",
                "longName":"Bitcoin",
                "currencyDecimalPlaces":8,
                "supportedWithdrawDecimalPlaces":8
             },
             "quoteCurrency":{
                "id":1,
                "symbol":"R",
                "decimalPlaces":2,
                "isActive":true,
                "shortName":"ZAR",
                "longName":"Rand",
                "currencyDecimalPlaces":2,
                "supportedWithdrawDecimalPlaces":2
             },
             "shortName":"BTC/ZAR",
             "exchange":"VALR",
             "active":true,
             "minBaseAmount":0.0001,
             "maxBaseAmount":2,
             "minQuoteAmount":10,
             "maxQuoteAmount":100000
          },
          "originalPrice":"80000",
          "remainingQuantity":"0.01",
          "originalQuantity":"0.01",
          "orderSide":"buy",
          "orderType":"limit",
          "failedReason":"",
          "orderUpdatedAt":"2019-05-10T14:47:24.826Z",
          "orderCreatedAt":"2019-05-10T14:42:37.333Z",
          "customerOrderId":"4"
       }
    }

    orderStatusType can be one of the following values: "Placed", "Failed", "Cancelled", "Filled", "Partially Filled",
    "Instant Order Balance Reserve Failed", "Instant Order Balance Reserved","Instant Order Completed".


    FAILED_CANCEL_ORDER : UNABLE TO CANCEL ORDER

    Sample message feed:

    {
       "type":"FAILED_CANCEL_ORDER",
       "data":{
          "orderId":"247dc157-bb5b-49af-b476-2f613b780697",
          "message":"An error occurred while cancelling your order."
       }
    }

    NEW_PENDING_RECEIVE : NEW PENDING CRYPTO DEPOSIT

    Sample message feed:

    {
       "type":"NEW_PENDING_RECEIVE",
       "data":{
          "currency":{
             "id":3,
             "symbol":"ETH",
             "decimalPlaces":8,
             "isActive":true,
             "shortName":"ETH",
             "longName":"Ethereum",
             "currencyDecimalPlaces":18,
             "supportedWithdrawDecimalPlaces":8
          },
          "receiveAddress":"0xA7Fae2Fd50886b962d46FF4280f595A3982aeAa5",
          "transactionHash":"0x804bbfa946b57fc5ffcb0c37ec02e7503435d19c35bf8eb0b0c6deb289f7009a",
          "amount":0.01,
          "createdAt":"2019-04-25T21:16:28Z",
          "confirmations":1,
          "confirmed":false
       }
    }

    This message feed is sent through every time there is an update to the number of confirmations to this
    pending deposit.


    SEND_STATUS_UPDATE : CRYPTO WITHDRAWAL STATUS UPDATE

    Sample message feed:

    {
       "type":"SEND_STATUS_UPDATE",
       "data":{
          "uniqueId":"beb8a612-1a1a-4d68-9bd3-96d5ea341119",
          "status":"SEND_BROADCASTED",
          "confirmations":0
       }
    }
    """

    _WEBSOCKET_API_URI = 'wss://api.valr.com'
    _ACCOUNT_CONNECTION = f'{_WEBSOCKET_API_URI}{WebSocketType.ACCOUNT.value}'
    _TRADE_CONNECTION = f'{_WEBSOCKET_API_URI}{WebSocketType.TRADE.value}'

    def __init__(self, api_key: str, api_secret: str, hooks: Dict[str, Callable],
                 currency_pairs: Optional[List[str]] = None, ws_type: str = 'trade',
                 trade_subscriptions: Optional[List[str]] = None):
        self._api_key = api_key
        self._api_secret = api_secret
        self._ws_type = WebSocketType[ws_type.upper()]
        self._hooks = {get_event_type(self._ws_type)[e.upper()]: f for e, f in hooks.items()}
        if currency_pairs:
            self._currency_pairs = [CurrencyPair[p.upper()] for p in currency_pairs]
        else:
            self._currency_pairs = [p for p in CurrencyPair]
        if self._ws_type == WebSocketType.ACCOUNT:
            self._uri = self._ACCOUNT_CONNECTION
        else:
            self._uri = self._TRADE_CONNECTION
        if self._ws_type == WebSocketType.TRADE:
            if trade_subscriptions:
                self._trade_subscriptions = [TradeEvent[e] for e in trade_subscriptions]
            else:
                self._trade_subscriptions = [e for e in TradeEvent]
        elif trade_subscriptions:
            raise ValueError(f'trade subscriptions requires ws_type of {WebSocketType.TRADE.name} ')
        else:
            self._trade_subscriptions = None

    async def run(self):
        """Open an async websocket connection, consume responses and executed mapped hooks.  Async hooks are also
        supported.  The method relies on the underlying 'websockets' libraries ping-pong support.  No API-level
        ping-pong messages are sent to keep the connection alive (not necessary).  Support for custom-handling of
        websockets.exceptions.ConnectionClosed must be handled in the application.
        """
        headers = _get_valr_headers(api_key=self._api_key, api_secret=self._api_secret, method='GET',
                                    path=self._ws_type.value, data='')
        async with websockets.connect(self._uri, ssl=True, extra_headers=headers) as ws:
            if self._ws_type == WebSocketType.TRADE:
                await ws.send(self.get_subscribe_data(self._currency_pairs, self._trade_subscriptions))
            async for message in ws:
                data = json.loads(message)
                try:
                    # ignore auth and subscription response messages
                    if data['type'] not in (MessageFeedType.SUBSCRIBED.name, MessageFeedType.AUTHENTICATED.name):
                        func = self._hooks[get_event_type(self._ws_type)[data['type']]]
                        # apply hooks to mapped stream events
                        if asyncio.iscoroutinefunction(func):
                            await func(data)
                        else:
                            func(data)
                except KeyError:
                    events = [e.name for e in get_event_type(self._ws_type)]
                    if data['type'] in events:
                        raise HookNotFoundError(f'no hook supplied for {data["type"]} event')
                    raise WebSocketAPIException(f'WebSocket API failed to handle {data["type"]} event: {data}')

    @staticmethod
    def get_subscribe_data(currency_pairs, events) -> JSONType:
        """Get subscription data for ws client request"""
        subscriptions = [{"event": e.name, "pairs": [p.name for p in currency_pairs]} for e in events]
        data = {
            "type": MessageFeedType.SUBSCRIBE.name,
            "subscriptions": subscriptions
        }
        return json.dumps(data, default=str)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
    name='valr-python',
    version='0.2.0',
    license='MIT',
    description='Python SDK for the VALR REST API',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Jonathan Els',
    author_email='jonathanelscpt@gmail.com',
    url='https://github.com/jonathanelscpt/valr-python',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    project_urls={
        'Changelog': 'https://github.com/jonathanelscpt/valr-python/blob/master/CHANGELOG.rst',
        'Issue Tracker': 'https://github.com/jonathanelscpt/valr-python/issues',
    },
    keywords=[
        'VALR', 'REST', 'API', 'Bitcoin', 'Ethereum', 'stream', 'websocket',
    ],
    python_requires='>=3.6',
    install_requires=[
        'requests', 'websockets', 'asyncio',
    ],
    tests_require=['pytest', 'pytest-cov', 'requests_mock'],
    extras_require={
    },
)



from time import sleep

LIVE_API_TIMEOUT = 4


def teardown_function(function):
    """Add timeout between live public api calls to avoid HTTP 429 rate-limiting"""
    sleep(LIVE_API_TIMEOUT)


def test_live_get_order_book_public(sync_client, btc_zar):
    resp = sync_client.get_order_book_public(currency_pair=btc_zar)
    assert isinstance(resp, dict)
    assert 'Asks' in resp and 'Bids' in resp
    assert isinstance(resp['Asks'], list) and isinstance(resp['Bids'], list)


def test_live_get_currencies(sync_client):
    resp = sync_client.get_currencies()
    assert isinstance(resp, list)
    assert isinstance(resp[0], dict)


def test_live_get_currency_pairs(sync_client):
    resp = sync_client.get_currency_pairs()
    assert isinstance(resp, list)
    assert isinstance(resp[0], dict)


def test_live_get_order_types_no_currency_pair(sync_client):
    resp = sync_client.get_order_types()
    assert isinstance(resp, list)
    assert isinstance(resp[0], dict)


def test_live_get_order_types_with_currency_pair(sync_client, btc_zar):
    resp = sync_client.get_order_types(currency_pair=btc_zar)
    assert isinstance(resp, list)
    assert isinstance(resp[0], str)


def test_live_get_market_summary_no_currency_pair(sync_client):
    resp = sync_client.get_market_summary()
    assert isinstance(resp, list)
    assert isinstance(resp[0], dict)


def test_live_get_market_summary_with_currency_pair(sync_client, btc_zar):
    resp = sync_client.get_market_summary(currency_pair=btc_zar)
    assert isinstance(resp, dict)


def test_live_get_server_time(sync_client):
    resp = sync_client.get_server_time()
    assert isinstance(resp, dict)



from valr_python.rest_base import BaseClientABC

BASE_URL = BaseClientABC.VALR_API_URL


# Public APIs

def test_get_order_book_public(rest_sync_mocker, sync_client, rest_sync_mock_resp, btc_zar):
    rest_sync_mocker.get(f'{BASE_URL}/v1/public/{btc_zar}/orderbook', json=rest_sync_mock_resp)
    sdk_resp = sync_client.get_order_book_public(btc_zar)
    assert sdk_resp == rest_sync_mock_resp


def test_get_currencies(rest_sync_mocker, sync_client, rest_sync_mock_resp):
    rest_sync_mocker.get(f'{BASE_URL}/v1/public/currencies', json=rest_sync_mock_resp)
    sdk_resp = sync_client.get_currencies()
    assert sdk_resp == rest_sync_mock_resp


def test_get_currency_pairs(rest_sync_mocker, sync_client, rest_sync_mock_resp):
    rest_sync_mocker.get(f'{BASE_URL}/v1/public/pairs', json=rest_sync_mock_resp)
    sdk_resp = sync_client.get_currency_pairs()
    assert sdk_resp == rest_sync_mock_resp


def test_get_order_types(rest_sync_mocker, sync_client, rest_sync_mock_resp, btc_zar):
    rest_sync_mocker.get(f'{BASE_URL}/v1/public/ordertypes', json=rest_sync_mock_resp)
    sdk_resp_no_currency_pair = sync_client.get_order_types()
    assert sdk_resp_no_currency_pair == rest_sync_mock_resp

    rest_sync_mocker.get(f'{BASE_URL}/v1/public/{btc_zar}/ordertypes', json=rest_sync_mock_resp)
    sdk_resp = sync_client.get_order_types(currency_pair=btc_zar)
    assert sdk_resp == rest_sync_mock_resp


def test_get_market_summary(rest_sync_mocker, sync_client, rest_sync_mock_resp, btc_zar):
    rest_sync_mocker.get(f'{BASE_URL}/v1/public/marketsummary', json=rest_sync_mock_resp)
    sdk_resp_no_currency_pair = sync_client.get_market_summary()
    assert sdk_resp_no_currency_pair == rest_sync_mock_resp

    rest_sync_mocker.get(f'{BASE_URL}/v1/public/{btc_zar}/marketsummary', json=rest_sync_mock_resp)
    sdk_resp_no_currency_pair = sync_client.get_market_summary()
    assert sdk_resp_no_currency_pair == rest_sync_mock_resp


def test_get_server_time(rest_sync_mocker, sync_client, rest_sync_mock_resp):
    rest_sync_mocker.get(f'{BASE_URL}/v1/public/time', json=rest_sync_mock_resp)
    sdk_resp = sync_client.get_server_time()
    assert sdk_resp == rest_sync_mock_resp



import pytest

from valr_python.exceptions import RequiresAuthentication
from valr_python.rest_base import BaseClientABC

BASE_URL = BaseClientABC.VALR_API_URL


# Account APIs

def test_get_balances(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp):
    rest_sync_mocker.get(f'{BASE_URL}/v1/account/balances', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_balances()

    sdk_resp = sync_client_with_auth.get_balances()
    assert sdk_resp == rest_sync_mock_resp


def test_get_transaction_history(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp):
    skip = 5
    limit = 5
    rest_sync_mocker.get(f'{BASE_URL}/v1/account/transactionhistory?skip={skip}&limit={limit}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_transaction_history(skip=skip, limit=limit)

    sdk_resp = sync_client_with_auth.get_transaction_history(skip=skip, limit=limit)
    assert sdk_resp == rest_sync_mock_resp


def test_get_account_trade_history(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    limit = 5
    rest_sync_mocker.get(f'{BASE_URL}/v1/account/{btc_zar}/tradehistory?limit={limit}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_account_trade_history(currency_pair=btc_zar, limit=limit)

    sdk_resp = sync_client_with_auth.get_account_trade_history(currency_pair=btc_zar, limit=limit)
    assert sdk_resp == rest_sync_mock_resp


# Crypto Wallet APIs

def test_get_deposit_address(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc):
    rest_sync_mocker.get(f'{BASE_URL}/v1/wallet/crypto/{btc}/deposit/address', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_deposit_address(currency_code=btc)

    sdk_resp = sync_client_with_auth.get_deposit_address(currency_code=btc)
    assert sdk_resp == rest_sync_mock_resp


def test_get_withdrawal_info(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc):
    rest_sync_mocker.get(f'{BASE_URL}/v1/wallet/crypto/{btc}/withdraw', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_withdrawal_info(currency_code=btc)

    sdk_resp = sync_client_with_auth.get_withdrawal_info(currency_code=btc)
    assert sdk_resp == rest_sync_mock_resp


def test_post_new_crypto_withdrawal(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc):
    amount = 0.001
    address = 'crypto-address'
    payment_reference = 'payment-ref'
    rest_sync_mocker.post(f'{BASE_URL}/v1/wallet/crypto/{btc}/withdraw', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.post_new_crypto_withdrawal(currency_code=btc, amount=amount, address=address,
                                               payment_reference=payment_reference)

    sdk_resp = sync_client_with_auth.post_new_crypto_withdrawal(currency_code=btc, amount=amount, address=address,
                                                                payment_reference=payment_reference)
    assert sdk_resp == rest_sync_mock_resp


def test_get_withdrawal_status(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc):
    withdraw_id = 'withdraw_id'
    rest_sync_mocker.get(f'{BASE_URL}/v1/wallet/crypto/{btc}/withdraw/{withdraw_id}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_withdrawal_status(currency_code=btc)

    sdk_resp = sync_client_with_auth.get_withdrawal_status(currency_code=btc, withdraw_id=withdraw_id)
    assert sdk_resp == rest_sync_mock_resp


def test_get_deposit_history(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc):
    skip = 5
    limit = 5
    rest_sync_mocker.get(f'{BASE_URL}/v1/wallet/crypto/{btc}/deposit/history?skip={skip}&limit={limit}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_deposit_history(currency_code=btc, skip=skip, limit=limit)

    sdk_resp = sync_client_with_auth.get_deposit_history(currency_code=btc, skip=skip, limit=limit)
    assert sdk_resp == rest_sync_mock_resp


def test_get_withdrawal_history(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc):
    skip = 5
    limit = 5
    rest_sync_mocker.get(f'{BASE_URL}/v1/wallet/crypto/{btc}/withdraw/history?skip={skip}&limit={limit}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_withdrawal_history(currency_code=btc, skip=skip, limit=limit)

    sdk_resp = sync_client_with_auth.get_withdrawal_history(currency_code=btc, skip=skip, limit=limit)
    assert sdk_resp == rest_sync_mock_resp


# Fiat Wallet APIs

def test_get_bank_accounts(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, zar):
    rest_sync_mocker.get(f'{BASE_URL}/v1/wallet/fiat/{zar}/accounts', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_bank_accounts(currency_code=zar)

    sdk_resp = sync_client_with_auth.get_bank_accounts(currency_code=zar)
    assert sdk_resp == rest_sync_mock_resp


def test_post_new_fiat_withdrawal(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, zar):
    linked_bank_account_id = 'linked_bank_account_id'
    amount = 100000.00
    rest_sync_mocker.post(f'{BASE_URL}/v1/wallet/fiat/{zar}/withdraw', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.post_new_fiat_withdrawal(currency_code=zar, amount=amount,
                                             linked_bank_account_id=linked_bank_account_id)

    sdk_resp = sync_client_with_auth.post_new_fiat_withdrawal(currency_code=zar, amount=amount,
                                                              linked_bank_account_id=linked_bank_account_id)
    assert sdk_resp == rest_sync_mock_resp


# Market Data APIs

def test_get_order_book(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    rest_sync_mocker.get(f'{BASE_URL}/v1/marketdata/{btc_zar}/orderbook', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_order_book(currency_pair=btc_zar)

    sdk_resp = sync_client_with_auth.get_order_book(currency_pair=btc_zar)
    assert sdk_resp == rest_sync_mock_resp


def test_get_order_book_full(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    rest_sync_mocker.get(f'{BASE_URL}/v1/marketdata/{btc_zar}/orderbook/full', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_order_book_full(currency_pair=btc_zar)

    sdk_resp = sync_client_with_auth.get_order_book_full(currency_pair=btc_zar)
    assert sdk_resp == rest_sync_mock_resp


def test_get_market_data_trade_history(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    limit = 5
    rest_sync_mocker.get(f'{BASE_URL}/v1/marketdata/{btc_zar}/tradehistory?limit={limit}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_market_data_trade_history(currency_pair=btc_zar, limit=5)

    sdk_resp = sync_client_with_auth.get_market_data_trade_history(currency_pair=btc_zar, limit=5)
    assert sdk_resp == rest_sync_mock_resp


# Simple Buy/Sell APIs

def test_post_simple_quote(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar, btc):
    pay_in_currency = btc
    pay_amount = 0.01
    side = "SELL"
    rest_sync_mocker.post(f'{BASE_URL}/v1/simple/{btc_zar}/quote', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.post_simple_quote(currency_pair=btc_zar, pay_in_currency=pay_in_currency,
                                      pay_amount=pay_amount, side=side)

    sdk_resp = sync_client_with_auth.post_simple_quote(currency_pair=btc_zar, pay_in_currency=pay_in_currency,
                                                       pay_amount=pay_amount, side=side)
    assert sdk_resp == rest_sync_mock_resp


def test_post_simple_order(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar, btc):
    pay_in_currency = btc
    pay_amount = 0.01
    side = "SELL"
    rest_sync_mocker.post(f'{BASE_URL}/v1/simple/{btc_zar}/order', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.post_simple_order(currency_pair=btc_zar, pay_in_currency=pay_in_currency,
                                      pay_amount=pay_amount, side=side)

    sdk_resp = sync_client_with_auth.post_simple_order(currency_pair=btc_zar, pay_in_currency=pay_in_currency,
                                                       pay_amount=pay_amount, side=side)
    assert sdk_resp == rest_sync_mock_resp


def test_get_simple_order_status(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    order_id = 'order_id'
    rest_sync_mocker.get(f'{BASE_URL}/v1/simple/{btc_zar}/order/{order_id}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_simple_order_status(currency_pair=btc_zar, order_id=order_id)

    sdk_resp = sync_client_with_auth.get_simple_order_status(currency_pair=btc_zar, order_id=order_id)
    assert sdk_resp == rest_sync_mock_resp


# Exchange Buy/Sell APIs

def test_post_limit_order(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    side = "SELL"
    quantity = 0.100000
    price = 123000
    post_only = True
    customer_order_id = 'customer_order_id'
    rest_sync_mocker.post(f'{BASE_URL}/v1/orders/limit', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.post_limit_order(side=side, quantity=quantity, price=price, pair=btc_zar, post_only=post_only,
                                     customer_order_id=customer_order_id)

    sdk_resp = sync_client_with_auth.post_limit_order(side=side, quantity=quantity, price=price, pair=btc_zar,
                                                      post_only=post_only, customer_order_id=customer_order_id)
    assert sdk_resp == rest_sync_mock_resp


def test_post_market_order(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    side = "SELL"
    base_amount = 0.100000
    quote_amount = 123000
    customer_order_id = 'customer_order_id'
    rest_sync_mocker.post(f'{BASE_URL}/v1/orders/market', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.post_market_order(side=side, pair=btc_zar, base_amount=base_amount,
                                      customer_order_id=customer_order_id)

    with pytest.raises(AttributeError):
        sync_client_with_auth.post_market_order(side=side, pair=btc_zar, base_amount=base_amount,
                                                quote_amount=quote_amount, customer_order_id=customer_order_id)

    sdk_resp = sync_client_with_auth.post_market_order(side=side, pair=btc_zar, base_amount=base_amount,
                                                       customer_order_id=customer_order_id)
    assert sdk_resp == rest_sync_mock_resp


def test_get_order_status(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    customer_order_id = "customer_order_id"
    order_id = "order_id"

    rest_sync_mocker.get(f'{BASE_URL}/v1/orders/{btc_zar}/orderid/{order_id}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_order_status(currency_pair=btc_zar, order_id=order_id)

    with pytest.raises(AttributeError):
        sync_client_with_auth.get_order_status(currency_pair=btc_zar, order_id=order_id,
                                               customer_order_id=customer_order_id)

    sdk_resp_order_id = sync_client_with_auth.get_order_status(currency_pair=btc_zar, order_id=order_id)
    assert sdk_resp_order_id == rest_sync_mock_resp

    rest_sync_mocker.get(f'{BASE_URL}/v1/orders/{btc_zar}/customerorderid/{customer_order_id}', json=rest_sync_mock_resp)
    sdk_resp_customer_order_id = sync_client_with_auth.get_order_status(currency_pair=btc_zar,
                                                                        customer_order_id=customer_order_id)
    assert sdk_resp_customer_order_id == rest_sync_mock_resp


def test_get_all_open_orders(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp):
    rest_sync_mocker.get(f'{BASE_URL}/v1/orders/open', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_all_open_orders()

    sdk_resp = sync_client_with_auth.get_all_open_orders()
    assert sdk_resp == rest_sync_mock_resp


def test_get_order_history(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp):
    skip = 5
    limit = 5
    rest_sync_mocker.get(f'/v1/orders/history?skip={skip}&limit={limit}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_order_history(skip=skip, limit=limit)

    sdk_resp = sync_client_with_auth.get_order_history(skip=skip, limit=limit)
    assert sdk_resp == rest_sync_mock_resp


def test_get_order_history_summary(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp):
    customer_order_id = "customer_order_id"
    order_id = "order_id"

    rest_sync_mocker.get(f'{BASE_URL}/v1/orders/history/summary/orderid/{order_id}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_order_history_summary(order_id=order_id)

    with pytest.raises(AttributeError):
        sync_client_with_auth.get_order_history_summary(order_id=order_id, customer_order_id=customer_order_id)

    sdk_resp_order_id = sync_client_with_auth.get_order_history_summary(order_id=order_id)
    assert sdk_resp_order_id == rest_sync_mock_resp

    rest_sync_mocker.get(f'{BASE_URL}/v1/orders/history/summary/customerorderid/{customer_order_id}', json=rest_sync_mock_resp)
    sdk_resp_customer_order_id = sync_client_with_auth.get_order_history_summary(customer_order_id=customer_order_id)
    assert sdk_resp_customer_order_id == rest_sync_mock_resp


def test_get_order_history_detail(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp):
    customer_order_id = "customer_order_id"
    order_id = "order_id"

    rest_sync_mocker.get(f'{BASE_URL}/v1/orders/history/detail/orderid/{order_id}', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.get_order_history_detail(order_id=order_id)

    with pytest.raises(AttributeError):
        sync_client_with_auth.get_order_history_detail(order_id=order_id, customer_order_id=customer_order_id)

    sdk_resp_order_id = sync_client_with_auth.get_order_history_detail(order_id=order_id)
    assert sdk_resp_order_id == rest_sync_mock_resp

    rest_sync_mocker.get(f'{BASE_URL}/v1/orders/history/detail/customerorderid/{customer_order_id}', json=rest_sync_mock_resp)
    sdk_resp_customer_order_id = sync_client_with_auth.get_order_history_detail(customer_order_id=customer_order_id)
    assert sdk_resp_customer_order_id == rest_sync_mock_resp


def test_delete_order(rest_sync_mocker, sync_client, sync_client_with_auth, rest_sync_mock_resp, btc_zar):
    customer_order_id = "customer_order_id"
    order_id = "order_id"

    rest_sync_mocker.delete(f'{BASE_URL}/v1/orders/order', json=rest_sync_mock_resp)

    with pytest.raises(RequiresAuthentication):
        sync_client.delete_order(pair=btc_zar, order_id=order_id)

    with pytest.raises(AttributeError):
        sync_client_with_auth.delete_order(pair=btc_zar, order_id=order_id, customer_order_id=customer_order_id)

    sdk_resp_order_id = sync_client_with_auth.delete_order(pair=btc_zar, order_id=order_id)
    assert sdk_resp_order_id == rest_sync_mock_resp

    sdk_resp_customer_order_id = sync_client_with_auth.delete_order(pair=btc_zar, customer_order_id=customer_order_id)
    assert sdk_resp_customer_order_id == rest_sync_mock_resp



import pytest
from requests.exceptions import HTTPError

from valr_python.exceptions import APIError
from valr_python.exceptions import IncompleteOrderWarning
from valr_python.exceptions import RequiresAuthentication
from valr_python.exceptions import RESTAPIException
from valr_python.exceptions import TooManyRequestsWarning


def test_client_attrs(sync_client):
    sync_client.api_secret = 'api_secret'
    sync_client.api_key = 'api_key'
    sync_client.base_url = 'base_url'
    sync_client.timeout = 10
    sync_client.rate_limiting_support = True

    assert sync_client.api_key == 'api_key'
    assert sync_client.api_secret == 'api_secret'
    assert sync_client.base_url == 'base_url'
    assert sync_client.timeout == 10
    assert sync_client.rate_limiting_support is True


def test_client_do_basic(mock_sync_client, rest_sync_mocker):
    rest_sync_mocker.get('mock://test/', json={"key": "value"}, status_code=200)

    # valid k/v responses
    res = mock_sync_client._do('GET', '/')
    assert res['key'] == 'value'


def test_client_do_authentication_success(mock_sync_client, rest_sync_mocker, rest_sync_mock_resp):
    rest_sync_mocker.get('mock://test/', json=rest_sync_mock_resp)
    mock_sync_client.api_secret = 'api_secret'
    mock_sync_client.api_key = 'api_key'
    resp = mock_sync_client._do('GET', '/', is_authenticated=True)
    assert resp == rest_sync_mock_resp


def test_client_do_authentication_no_key_secret_pair(mock_sync_client, rest_sync_mocker, rest_sync_mock_resp):
    rest_sync_mocker.get('mock://test/', json=rest_sync_mock_resp)
    with pytest.raises(RequiresAuthentication):
        # fail as no api key/secret
        mock_sync_client._do('GET', '/', is_authenticated=True)


def test_client_do_api_error_handling(mock_sync_client, rest_sync_mocker):
    rest_sync_mocker.get('mock://test/', json={"code": "-12345", "message": "api error message"}, status_code=400)
    with pytest.raises(APIError) as e:
        mock_sync_client._do('GET', '/')
    assert e.value.code == '-12345'
    assert e.value.message == 'api error message'


def test_client_do_200_ok_error_handling(mock_sync_client, rest_sync_mocker):
    rest_sync_mocker.get('mock://test/', json={"code": "-12345", "message": "api error message"}, status_code=200)
    with pytest.raises(APIError) as e:
        mock_sync_client._do('GET', '/')
    assert e.value.code == '-12345'
    assert e.value.message == 'api error message'


def test_client_do_warn_on_202_response(mock_sync_client, rest_sync_mocker):
    _202_resp = {"id": "order-id"}
    rest_sync_mocker.get('mock://test/', json=_202_resp, status_code=202)
    with pytest.warns(IncompleteOrderWarning) as w:
        resp = mock_sync_client._do('GET', '/')
        assert resp == _202_resp
        # check that warning bundles expected response data
        assert w[0].message.data == _202_resp


def test_client_do_invalid_response_handling(mock_sync_client, rest_sync_mocker):
    rest_sync_mocker.get('mock://test/', text='invalid json response')
    with pytest.raises(RESTAPIException):
        mock_sync_client._do('GET', '/')


def test_client_do_http_error_handling(mock_sync_client, rest_sync_mocker):
    # HTTP errors without VALR api handling
    rest_sync_mocker.get('mock://test/', json={'error': 'Internal Server Error'}, status_code=500)
    with pytest.raises(HTTPError):
        mock_sync_client._do('GET', '/')


def test_client_do_http_429_handling(mock_sync_client, rest_sync_mocker):
    _429_resp = {'status_code': 429, "headers": {"Retry-After": "1"}}
    _200_resp = {'json': {"key": "value"}, "status_code": 200}
    resp_list = [_429_resp, _200_resp]

    # fail without 429 handling flag set
    rest_sync_mocker.get('mock://test/', resp_list)
    with pytest.raises(HTTPError):
        mock_sync_client._do('GET', '/')

    # handle 429s when enabled and validate warning issued
    with pytest.warns(TooManyRequestsWarning):
        mock_sync_client.rate_limiting_support = True
        rest_sync_mocker.get('mock://test/', resp_list)
        res = mock_sync_client._do('GET', '/')
        assert res['key'] == 'value'


@pytest.mark.parametrize('headers', [{}, {"Retry-After": "bogus"}])
def test_client_do_http_429_api_exception_handling(mock_sync_client, rest_sync_mocker, headers):
    mock_sync_client.rate_limiting_support = True

    # raise api exception if header not in response
    rest_sync_mocker.get('mock://test/', headers={}, status_code=429)
    with pytest.raises(RESTAPIException):
        mock_sync_client._do('GET', '/')

    # raise api exception if "Retry-After" parsing fails
    rest_sync_mocker.get('mock://test/', headers={"Retry-After": "bogus"}, status_code=429)
    with pytest.raises(RESTAPIException):
        mock_sync_client._do('GET', '/')



import pytest

from valr_python.decorators import check_xor_attrs
from valr_python.decorators import requires_authentication
from valr_python.exceptions import RequiresAuthentication


class DecoratorStub(object):
    def __init__(self, api_key=None, api_secret=None):
        self._api_key = api_key
        self._api_secret = api_secret

    @requires_authentication
    def private_action(self):
        return True

    @check_xor_attrs('attr1', 'attr2')
    def xor_function(self, attr1=None, attr2=None):  # noqa
        return True


@pytest.mark.parametrize('api_key, api_secret',
                         [('api_key', None), (None, 'api_secret'), (None, None), ("", "")])
def test_requires_authentication_failures(api_key, api_secret):
    stub = DecoratorStub(api_secret=api_secret, api_key=api_key)
    with pytest.raises(RequiresAuthentication):
        stub.private_action()


@pytest.mark.parametrize('api_key, api_secret', [('api_key', ' api_secret')])
def test_requires_authentication_successful(api_key, api_secret):
    stub = DecoratorStub(api_secret=api_secret, api_key=api_key)
    assert stub.private_action() is True


@pytest.mark.parametrize('attr1, attr2', [(True, True)])
def test_check_xor_attrs_failures(attr1, attr2):
    stub = DecoratorStub()
    with pytest.raises(AttributeError):
        stub.xor_function(attr1=attr1, attr2=attr2)
    with pytest.raises(AttributeError):
        stub.xor_function()


@pytest.mark.parametrize('attr1, attr2', [(True, True)])
def test_check_xor_attrs_successful(attr1, attr2):
    stub = DecoratorStub()
    assert stub.xor_function(attr1=attr1) is True
    assert stub.xor_function(attr2=attr2) is True
